"""Regras regex para ato de Nomeacao de Comissionados."""

import os
import joblib
from dodfminer.extract.polished.acts.base import Atos


class NomeacaoComissionados(Atos):

    def __init__(self, file, backend, segmentation):
        super().__init__(file, backend)

    def _act_name(self):
        return "Nomeação"

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/nomeacao_ner.pkl'
        return joblib.load(f_path)

    def _props_names(self):
        return ['Tipo do Ato', 'Nome', 'Cargo Efetivo', 'Matricula', 'Siape',
                'Simbolo', 'Cargo Comissao', 'Lotacao', 'Orgao']

    def _rule_for_inst(self):
        start = r"(NOMEAR)"
        body = r"([\s\S]*?)"
        end = r"\.\s"
        return start + body + end

    def _prop_rules(self):
        eft = r"(?:^[A-ZÀ-Ž\s]+[A-ZÀ-Ž]),\s(?![M|m]atr[i|í]cula)([\s\S]*?),\s"
        siape = r"[S|s][I|i][A|a][P|p][E|e]\s[N|n]?[o|O]?\s([\s\S]*?)[,| | .]"
        coms = r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])\sde([\s\S]*?),"
        lot = r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])"
        lot2 = r"\sde(?:[\s\S]*?),\sd[a|e|o]\s([\s\S]*,?),"
        org = r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])\sde(?:[\s\S]*?)"
        org2 = r",\sd[a|e|o]\s(?:[\s\S]*,?),\sd[a|e|o]\s([\s\S]*?)$"

        rules = {"nome": r"(^[A-ZÀ-Ž\s]+[A-ZÀ-Ž])",
                 "cargo_efetivo": eft,
                 "matricula": r"[M|m]atr[í|i]cula\s?n?o?\s([\s\S]*?)[,|\s]",
                 "matricula_SIAPE": siape,
                 "simbolo": r"[S|s][í|i]mbolo\s?n?o?\s([\s\S]*?)[,|\s]",
                 "cargo_comissionado": coms,
                 "hierarquia_lotacao": lot + lot2,
                 "orgao": org + org2}
        return rules

class NomeacaoEfetivos(Atos):

    def __init__(self, file, backend):
        super().__init__(file, backend)

    def _act_name(self):
        return "Nomeação de Efetivos"

    def _props_names(self):
        return ['tipo','edital_normativo','data_do_edital_normativo','DODF_edital_normativo','data_DODF_edital_normativo','edital_resultado_final','data_edital_resultado_final','cargo','especialiade','carreira','orgao','nome_candidato','classificacao','pne','sei','reposicionamento']

    def _rule_for_inst(self):
        start = r"(NOMEAR\s)((?:[ao]s\scandidat[ao]s\sabaixo(?:[a-zA-Z_0-9,\s\/-\:\-\(\);]*).|(?:[ao]\scandidat[oa]\sabaixo(?:[a-zA-Z_0-9,\s\/-\:\-\(\)]*)))(?:\s[a-zA-Z_\s]*(?:deficiencia|especiais):(?:\s[\sA-Zo]+,\s?\d{1,4}o?;?)+)?(?:\s)?(?:[\r\n\t\f\sa-zA-Z_\s]*classificacao:(?:\s[\sA-Zo]+,\s?\d{1,4}o?[,;]?)+)?)"
        body = ""
        end = ""
        return start + body + end

    def _prop_rules(self):
        rules = rules = {"edital_normativo": r"Edital\s(?:[Nn]ormativo|de\s[Aa]bertura)\sno\s([\/\s\-a-zA-Z0-9_]+)",
                         "data_edital_normativo": r"",
                         "numero_dodf_edital_normativo": r"publicado\sno\sDODF\sno\s(\d{1,3})",
                         "data_dodf_edital_normativo": r"publicado\sno\sDODF\sno\s\d{1,3},\s?de([\s0-9a-or-vzç]*\d{4})",
                         "edital_resultado_final": r"Resultado\sFinal\sno\s([\/\s\-a-zA-Z0-9_]+)",
                         "data_edital_resultado_final": r"",
                         "cargo": r"DODF\sno\s\d{1,3}(?:,[\s0-9a-or-vzç]*\d{4}),[\sa-z]*([A-Z\s]+)",
                         "especialiade": r"[\s,a-z\(\:\)]+([\sA-Z\-]*):",
                         "carreira": r"[cC]arreira\s(?:d[ae]\s)?([a-zA-Z\s]+)",
                         "orgao": r"[cC]arreira\s(?:d[ae]\s)?(?:[a-zA-Z\s]+),\s?d?[ao]?\s?([\sa-zA-Z0-9_]*)",
                         "candidato": r"(?:[\sA-Z\-]*):(?:[\sa-zC]*:)?\s([\sA-Z0-9\,o\;]+)",
                         "classificacao": r"",
                         "pne": r"(?:deficiencia|especiais):\s([\sA-Z0-9\,o\;]+)",
                         "processo_SEI": r"(?<!lei)\s((?:[0-9|\s]*?[.|-]\s?)+?[0-9|\s]*/\s?[0-9|\s]*-?\s?[0-9|\s]*)[.|,]",
                         "reposicionamento": r"lista\sde\sclassificacao:\s([\sA-Z0-9\,o\;]+)"}
        return rules



import torch
import os
from torch import nn
import numpy as np
import torch
import nltk
from nltk.tokenize import sent_tokenize
from torch.nn.utils.rnn import pad_sequence 
import unicodedata
import joblib
import gdown

model_cover_list = ['Aposentadoria']


class word_cnn(nn.Module):
    """single layer CNN with: filters=50, kernel_size=3, dropout=0.5"""
    def __init__(self, embedding_dim, embedding_size=None, pretrained_word_emb=None, pad_idx = 929606):
        super(word_cnn, self).__init__()
        assert (embedding_size or pretrained_word_emb), 'Undefined embedding for the word level CNN'

        if pretrained_word_emb:
            self.embedding = nn.Embedding.from_pretrained(torch.FloatTensor(pretrained_word_emb.vectors))
            self.embedding.padding_idx = pad_idx
        else:
            self.embedding = nn.Embedding(num_embeddings=embedding_size, embedding_dim=embedding_dim, padding_idx=pad_idx)

        self.conv = nn.Conv1d(in_channels=embedding_dim, out_channels=200, kernel_size=3, stride=1, padding=2)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.5)

    def forward(self, x):
        x = self.embedding(x)
        shape = x.shape
        x = self.conv(x.view([shape[0]*shape[1], shape[2], shape[3]]).permute(0, 2, 1))
        x = self.dropout(self.relu(x))
        x = torch.nn.functional.max_pool1d(x, kernel_size=x.shape[2]).squeeze()
        return x.view([shape[0], shape[1], -1])

class sentence_cnn(nn.Module):
    """two-layer CNN, filter=800, kernel=5"""
    def __init__(self, in_channels=200):
        super(sentence_cnn, self).__init__()
        self.conv1 = nn.Conv1d(in_channels=in_channels, out_channels=800, kernel_size=5, padding=2)
        self.conv2 = nn.Conv1d(in_channels=800, out_channels=800, kernel_size=5, padding=2)
        self.relu  = nn.ReLU()
        self.dropout = nn.Dropout(p=0.5)

    def forward(self, x):
        w = x.clone()
        # Go through 2 conv layers
        x = self.conv1(x.permute(0, 2, 1))
        x = self.dropout(self.relu(x))
        x = self.conv2(x)
        x = self.relu(x)
        x = x.permute(0, 2, 1)
        return torch.cat((x, w), dim=2)

class decoder(nn.Module):
    def __init__(self, feature_size, num_classes, device):
        super(decoder, self).__init__()
        self.device = device
        self.num_classes = num_classes
        self.lstm = torch.nn.LSTM(feature_size+num_classes, 256)
        self.linear = torch.nn.Linear(256, num_classes)
        self.dropout = torch.nn.Dropout(p=0.5)
        self.loss_fn = torch.nn.CrossEntropyLoss(ignore_index = -1, reduction='sum')
    
    def forward_step(self, x, prev_tag, prev_lstm_state):
        prev_tag_onehot = torch.nn.functional.one_hot(prev_tag, num_classes=self.num_classes).to(self.device)
        lstm_input = torch.cat((x, prev_tag_onehot), dim=1).unsqueeze(dim=0)
        lstm_output, lstm_state = self.lstm(lstm_input, prev_lstm_state)
        linear_output = self.linear(lstm_output.squeeze(dim=0))
        prediction = torch.argmax(linear_output, dim=1)
        return linear_output, prediction, lstm_state
    
    def forward(self, x, tag):
        batch_size, seq_len, _ = x.size()
        x = x.permute(1, 0, 2)
        tag = tag.permute(1, 0)
        pred = torch.LongTensor([0 for i in range(batch_size)])
        lstm_state = None
        loss = 0.0
        # 50% chance word mask (dropout) to improve generalization
        x = self.dropout(x)
        for i in range(seq_len):
            output, pred, lstm_state = self.forward_step(x=x[i], prev_tag=pred, prev_lstm_state=lstm_state)
            loss += self.loss_fn(output, tag[i])
        return loss

    def decode(self, x):
        batch_size, seq_len, _ = x.size()
        x = x.permute(1, 0, 2)
        pred = torch.LongTensor([0 for i in range(batch_size)])
        lstm_state = None
        full_predictions = torch.LongTensor().to(self.device)
        full_output = torch.Tensor().to(self.device)
        for i in range(seq_len):
            output, pred, lstm_state = self.forward_step(x=x[i], prev_tag=pred, prev_lstm_state=lstm_state)
            full_predictions = torch.cat((full_predictions, pred.unsqueeze(dim=0)), dim=0)
            full_output = torch.cat((full_output, output.unsqueeze(dim=0)), dim=0)
        return full_predictions.permute(1, 0), full_output.permute(1, 0, 2)


class CNN_CNN_LSTM(nn.Module):
    def __init__(self, embedding_dim, decoder_feature_size, num_classes, device, embedding_size=None, pretrained_word_emb=None):
        super(CNN_CNN_LSTM, self).__init__()
        assert embedding_size or pretrained_word_emb, 'Undefined word embedding :('
        self.num_classes = num_classes
        if pretrained_word_emb:
            self.word_encoder = word_cnn(pretrained_word_emb=pretrained_word_emb, embedding_dim=embedding_dim)
        else:
            self.word_encoder = word_cnn(embedding_dim=embedding_dim, embedding_size=embedding_size)
        self.sent_encoder = sentence_cnn()
        self.decoder      = decoder(num_classes=num_classes, feature_size=decoder_feature_size, device=device)

    def forward(self, x, y):
        x = self.word_encoder(x)
        x = self.sent_encoder(x)
        x = self.decoder(x, y)
        return x

    def decode(self, x):
        x = self.word_encoder(x)
        x = self.sent_encoder(x)
        predictions, output = self.decoder.decode(x)
        return predictions, output

### Metrics

def relaxed_f1_score(model, dataloader, device):
    TP = [0 for _ in range(len(model.num_classes))]
    FP = [0 for _ in range(len(model.num_classes))]
    FN = [0 for _ in range(len(model.num_classes))]

    # for sent, tag, word, mask in dataloader:
        # sent = sent.to(device)
        # tag = tag.to(device)
        # word = word.to(device)
        # pred, _ = model.eval().decode(sent, word)
    print("Still ongoing development!")

def exact_f1_score(model, dataloader, device):
    TP = np.array([0 for _ in range((model.num_classes-2)//2)])
    FP = np.array([0 for _ in range((model.num_classes-2)//2)])
    FN = np.array([0 for _ in range((model.num_classes-2)//2)])

    for x, y, mask in dataloader:
        x = x.to(device)
        y = y.to(device)
        pred, _ = model.eval().decode(x)

        batch_size = pred.shape[0]
        for i in range(batch_size):
            predicted_entities = find_entities(pred[i])
            real_entities = find_entities(y[i])
            for entity in predicted_entities:
                if entity in real_entities:
                    TP[(entity[2]//2)-1] += 1
                else:
                    FP[(entity[2]//2)-1] += 1
            for entity in real_entities:
                if entity not in predicted_entities:
                    FN[(entity[2]//2)-1] += 1

    precision = TP/(TP+FP+0.000001)
    recall = TP/(TP+FN+0.000001)
    f1 = 2*(precision*recall)/(precision+recall)

    occurrences = TP + FN
    return occurrences, f1

def find_entities(tag):
    entities = []
    prev_tag = 1
    begin_entity = -1

    for i in range(len(tag)):
        # Check if current tag is new entity by checking if it's 'B-' of any class
        if tag[i]%2==0 and tag[i]>=2:
            if prev_tag >=2:
                entities.append((begin_entity, i-1, prev_tag-1))
            begin_entity = i
            prev_tag = tag[i]+1
        # Check if current tag is new entity (by comparing to previous tag)
        elif tag[i] != prev_tag:
            if prev_tag >= 2:
                entities.append((begin_entity, i-1, prev_tag-1))
            begin_entity = i
            prev_tag = tag[i]
    # Check if entity continues to the end of tensor tag
    if prev_tag >=2:
        entities.append((begin_entity, len(tag)-1, prev_tag-1))
    return entities

### Interface

def load_model(word2idx_path, model_path):
    """
    Load the trained model and the dictionary to convert words into integers (to be fed to the neural model)
    """
    dic = joblib.load(word2idx_path)
    word2idx = dic['word2idx']
    embedding_size = dic['embedding_size']
    model = CNN_CNN_LSTM(embedding_dim=50, embedding_size=embedding_size, decoder_feature_size=1000, num_classes=4, device='cpu')
    model.load_state_dict(torch.load(model_path, map_location=torch.device('cpu')))
    return model, word2idx

def preprocess_dodf(text, word2idx_dict):
    """
    Preprocess a dodf pure text, splitting sentences 

    input: pure dodf extracted using DODFMiner (text file path)
    output: dodf split at sentence level and normalized (using unicodedata)
    """
    try:
        nltk.data.find('tokenizers/punkt')
    except LookupError:
        nltk.download('punkt')
    # Split text into sentences using NLTK sentence tokenizer
    sent_tokenizer=nltk.data.load('tokenizers/punkt/portuguese.pickle')
    sentences = sent_tokenizer.tokenize(text)
    processed_sentences = sentences.copy()
    processed_sentences = [[word.replace(',', '').replace(';', '').replace(':', '').lower() for word in sent.split()] for sent in processed_sentences]
    # processed_sentences = [sent.replace(',', '').replace(';', '').replace(':', '').split() for sent in processed_sentences]

    # Pad sentences to max_sentence_len
    max_sentence_len = -1
    for sent in processed_sentences:
        max_sentence_len = max(max_sentence_len, len(sent))
    for i in range(len(sentences)):
        processed_sentences[i] = ['<pad>' if k>=len(processed_sentences[i]) else processed_sentences[i][k] for k in range(max_sentence_len)]
    
    for i in range(len(processed_sentences)):
        for j in range(len(processed_sentences[i])):
            word = processed_sentences[i][j]
            if word in word2idx_dict:
                processed_sentences[i][j] = word2idx_dict[processed_sentences[i][j]]
            else:
                processed_sentences[i][j] = word2idx_dict["<unk>"]
    processed_sentences = torch.LongTensor(processed_sentences).unsqueeze(dim=0)
    return processed_sentences, sentences

def get_acts(predictions, sentences):
    """
    Extracts acts predicted by the text segmentation ML model

    input: predictions for each sentence, and the sentences that were given as input to the model (tuple indicating beginning, ending and type of an act)
    output: acts (one or more sentence per act) 
    """
    entities = find_entities(predictions)
    acts = []
    tags = []
    for prediction in entities:
        begin = prediction[0]
        end = prediction[1]
        tag = prediction[2]
        acts.append(' '.join(sentences[begin:end+1]))
        tags.append(tag)
    return acts, tags

#### Rotina

class Segmentation:
    @staticmethod
    def extract_segments(text, act_name):
        if act_name in model_cover_list:
            if not os.path.isfile(os.path.dirname(__file__) + '/model/cnn_cnn_lstm.pt'):
                print('[Segmentation] The model is not present in local files Downloding...')
                url = 'https://drive.google.com/uc?id=1keERcWl3M-ZMypqHWOtK3_TIntUI6I5s'
                output = os.path.dirname(__file__) + '/model/cnn_cnn_lstm.pt'
                gdown.download(url, output, quiet=False)
                print('[Segmentation] Download finished.')

            model, word2idx = load_model(word2idx_path=os.path.dirname(__file__) + '/model/word2idx_dictionary.pkl', 
                                         model_path=os.path.dirname(__file__) + '/model/cnn_cnn_lstm.pt')
            
            print('[Segmentation] Segmentation is being applied...')
            processed_sentences, sentences = preprocess_dodf(text=text, word2idx_dict=word2idx)
            predictions, _ = model.eval().decode(processed_sentences)
            atos, tags = get_acts(predictions.squeeze(dim=0), sentences)
            print('[Segmentation] Finished.')
            if len(atos) > 0:
                return atos[1]
            else:
                print('[Segmentation] No act found by segmentation returning the entire text')
                return text
        else:
            print('[Segmentation] The current model is not trained with this act yet. Changing to pure text')
            return text


"""Regras regex para ato de Exoneração."""

from dodfminer.extract.polished.acts.base import Atos


class Exoneracao(Atos):

    def __init__(self, file, backend, segmentation):
        super().__init__(file, backend)

    def _act_name(self):
        return "Exoneração"

    def _props_names(self):
        return ['Tipo do Ato', 'nome', 'matricula', 'simbolo', 'cargo_comissao',
                'lotacao', 'orgao', 'vigencia', 'pedido', 'cargo_efetivo',
                'siape', 'motivo']

    def _rule_for_inst(self):
        start = r"(EXONERAR)"
        body = r"([\s\S]*?)"
        end = r"\."
        return start + body + end

    def _prop_rules(self):
        siape = r"[S|s][I|i][A|a][P|p][E|e]\s[N|n]*[o|O]*\s?([\s\S]*?)[,| | .]"
        lot1 = r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])"
        lot2 = r"\sde(?:[\s\S]*?),\sd[a|e|o]\s([\s\S]*,?),"
        org = r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])\sde(?:[\s\S]*?)"
        org2 = r",\sd[a|e|o]\s(?:[\s\S]*,?),\sd[a|e|o]\s([\s\S]*?)$"
        rules = {"nome": r"([A-ZÀ-Ž\s]+[A-ZÀ-Ž])",
                 "matricula": r"matricula\s?n?o?\s([\s\S]*?-[\s\S]*?)[,]",
                 "simbolo": r"[S|s][í|i]mbolo\s?n?o?\s([\s\S]*?)[,|\s]",
                 "cargo_comissionado": r"(?:Cargo|Carreira)\sde([\s\S]*?)\,",
                 "hierarqui_lotacao": lot1 + lot2,
                 "orgao": org + org2,
                 "vigencia": r"",
                 "a_pedido_ou_nao": r"(a pedido)",
                 "cargo_efetivo": r"",
                 "matricula_SIAPE": siape,
                 "motivo": r""}

        return rules

class ExoneracaoEfetivos(Atos):

    def __init__(self, file, backend):
        super().__init__(file, backend)

    def _act_name(self):
        return "Exoneração Efetivos"

    def _find_instances(self):
        _instances = []
        pattern = r"([cC]omiss[aã]o|[nN]atureza\s?[eE]special)"
        _all = re.findall(self._inst_rule, self._text, flags=self._flags)
        it = re.finditer(self._inst_rule, self._text, flags=self._flags)
        for _ in it:
            _m = re.findall(pattern, _[0], 0)
            if not _m:
                _instances.append(_.groups())
        return _instances

    def _props_names(self):
        return ['tipo', 'nome','matricula','cargo_efetivo','classe','padrao','carreira','quadro', 'SEI','data','pedido', 'motivo', 'SIAPE', 'fundalamento_legal']

    def _rule_for_inst(self):
        start = r"(EXONERAR,\s?)"
        body = r"((?:a\spedido,)?\s(?:[A-Z\\n\s]+),\s(?:matr[ií]cula\s(?:[0-9\.-])+)[,\sa-zA-Z0-9\\\/-]*)"
        end = ""
        return start + body + end

    def _prop_rules(self):
        rules = {
                "nome": r"^(?:a\spedido,)?\s([A-Z\\n\s]+)",
                "matricula": r"matr[í|i]cula\s?n?o?\s([\s\S]*?)[,|\s]",
                "cargo_efetivo": r"matr[í|i]cula\s?n?o?\s[0-9]+,?([\sa-zA-Z]+)",
                "classe": r"matr[í|i]cula\s?n?o?\s[0-9]+,?[\sa-zA-Z]+[,\\n\s]+[eE]specialidade\s?([\sa-zA-Z]+)a\s?contar",
                "padrao": r"",
                "carreira": r"",
                "quadro": r"",
                "processo_SEI": r"SEI[a-z\s]*([0-9\-\/\n]+)",
                "data": r"a\scontar\sde\s([\s0-9\/]*)",
                "pedido": r"(a\spedido,)?\s(?:[A-Z\\n\s]+)",
                "motivo": r"",
                "matricula_SIAPE": r"",
                "fundamento_legal": r"nos\stermos\sdo[\n]?([a-zA-Z\s0-9\/]*)"}
        return rules



""" Polished extraction helper functions.

Functions in this files can be used inside, or outside, the ActsExtractor
class. Their purpose is to make some tasks easier for the user,
like creating txts, searching through files, and print dataframes.

Usage Example::

    from ddodfminer.extract.polished import helper
    helper.print_dataframe(df)

Functions
=========

"""

import os
import tqdm
import pandas as pd

from dodfminer.extract.polished.core import ActsExtractor
from dodfminer.extract.polished.core import _acts_ids
from dodfminer.extract.pure.core import ContentExtractor
from dodfminer.extract.polished.segmentation.core import Segmentation

def extract_segments(path):
    if os.path.isfile(path):
        files = [path]
    else:
        files = get_files_path(path, 'pdf')
    print("[Segmentation] Make yourself a coffee! This may take a while")
    bar = tqdm.tqdm(total=len(files), desc="[Segmentation] Progress")
    i = 1
    segments = ''
    for file in files:
        text = ContentExtractor.extract_text(file)
        # import pdb; pdb.set_trace()
        segments += '\n' +  Segmentation.extract_segments(text, 'Aposentadoria')
        i += 1
        bar.update(1)
    
    open('segmentation.txt', 'w+').write(segments)
        


def xml_multiple(path, backend):
    files = []
    if os.path.isfile(path):
        files = [path]
    else:
        files = get_files_path(path, 'pdf')

    print(files)
    print("[XMLFy] Make yourself a coffee! This may take a while")
    bar = tqdm.tqdm(total=len(files), desc="[XMLFy] Progress")
    i = 1
    for file in files:
        xml = ActsExtractor.get_xml(file, backend, i)
        xml.save_to_disc(path)
        i += 1
        bar.update(1)

def extract_multiple_acts(path, types, backend, segmentation=False):
    """Extract multple Acts from Multiple DODFs to act named CSVs.

    Args:
        path (str): Folder where the Dodfs are.
        types ([str]): Types of the act, see the core class to view
                    avaiables types.
        backend (str): what backend will be used to extract Acts {regex, ner}

    Returns:
        None
    """
    print(types)
    if len(types) == 0:
        types = _acts_ids.keys()

    if os.path.isfile(path):
        ContentExtractor.extract_text(path, single=True)
        for type in types:
            df = extract_single(path.replace('.pdf', '.txt'), type, backend=backend, segmentation=segmentation)
            df.to_csv(os.path.join(os.path.dirname(path), type+'.csv'))
    else:
        ContentExtractor.extract_to_txt(path)
        files = get_files_path(path, 'txt')
        for type in types:
            df = extract_multiple(files, type, backend, segmentation=segmentation)
            df.to_csv(os.path.join(path, type+".csv"))


def extract_multiple(files, type, backend, txt_out=False, txt_path="./results", segmentation=False):
    """Extract Act from Multiple DODF to a single DataFrame.

    Note:
        This function might save data to disc in text format,
        if txt_out is True.

    Args:
        files ([str]): List of dodfs files path.
        type (str): Type of the act, see the core class to view
                    avaiables types.
        backend (str): what backend will be used to extract Acts {regex, ner}
        txt_out (bool): Boolean indicating if acts should be saved on
                        text files.
        txt_path (str): Path to save the text files.

    Returns:
        A dataframe containing all instances of the desired
        act in the files set.

    """
    res = []
    for file in files:
        res_obj = ActsExtractor.get_act_obj(type, file, backend, segmentation)
        res_df = res_obj.data_frame
        res_txt = res_obj.acts_str
        if not res_df.empty:
            res.append(res_df)
            if txt_out:
                build_act_txt(res_txt, type, txt_path)

    if len(res) == 0:
        res_final = pd.DataFrame()
    else:
        res_final = pd.concat([pd.DataFrame(df) for df in res],
                              ignore_index=True)
    return res_final


def extract_single(file, type, backend, segmentation=False):
    """Extract Act from a single DODF to a single DataFrame.

    Note:
        This function might save data to disc in text format,
        if txt_out is True.

    Args:
        files (str): Dodf file path.
        type (str): Type of the act, see the core class to view
                    avaiables types.
        backend (str): what backend will be used to extract Acts {regex, ner}

    Returns:
        A dataframe containing all instances of the desired act
        including the texts found.

    """
    res_obj = ActsExtractor.get_act_obj(type, file, backend, segmentation)
    res_df = res_obj.data_frame
    res_txt = res_obj.acts_str
    res_df['text'] = res_txt

    return res_df


def build_act_txt(acts, name, save_path="./results/"):
    """Create a text file in disc for a act type.

    Note:
        This function might save data to disc in text format.

    Args:
        acts ([str]): List of all acts to save in the text file.
        name (str): Name of the output file.
        save_path (str): Path to save the text file.

    """
    if len(acts) > 0:
        file = open(f"{save_path}{name}.txt", "a")
        for act in acts:
            file.write(act)
            file.write("\n\n\n")
        file.close


def print_dataframe(df):
    """Style a Dataframe.

    Args:
        The dataframe to be styled.

    Returns:
        The styled dataframe

    """
    style_df = (df.style.set_properties(**{'text-align': 'left'})
                .set_table_styles([dict(selector='th',
                                        props=[('text-align', 'left')])]))
    return style_df


def get_files_path(path, type):
    """Get all files path inside a folder.

    Works with nested folders.

    Args:
        path: Folder to look into for files

    Returns:A dataframe containing all instances of the desired
        act in the files set.
        A list of strings with the file path.

    """
    files_path = []
    for root, _, files in os.walk(path):
        for file in files:
            if file.endswith("."+type):
                files_path.append(os.path.join(root, file))
    return files_path



"""Regras regex para ato de Aposentadoria."""

import re
import os
import joblib
from dodfminer.extract.polished.acts.base import Atos


class Retirements(Atos):

    def __init__(self, file, backend, segmentation):
        super().__init__(file, backend, segmentation)

    def _regex_flags(self):
        return re.IGNORECASE

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/aposentadoria.pkl'
        return joblib.load(f_path)

    def _act_name(self):
        return "Aposentadoria"

    def _props_names(self):
        return ["Tipo do Ato", "SEI", "Nome", "Matrícula",
                "Tipo de Aposentadoria", "Cargo", "Classe", "Padrao", "Quadro",
                "Fundamento Legal", "Orgao", "Vigencia", "Matricula SIAPE"]

    def _rule_for_inst(self):
        start = r"(APOSENTAR|CONCEDER\sAPOSENTADORIA,?\s?)"
        body = r"([\s\S]*?"
        end = r"(?<!lei)\s(?:[0-9|\s]*?[.|-]\s?)+?"
        end2 = r"[0-9|\s]*/\s?[0-9|\s]*-?\s?[0-9|\s]*[.|,])"
        return start + body + end + end2

    def _prop_rules(self):
        siape = r"[S|s][I|i][A|a][P|p][E|e]\s[N|n]?[o|O]?\s([\s\S]*?)[,| | .]"
        sei = r"(?<!lei)\s((?:[0-9|\s]*?[.|-]\s?)+?"
        sei2 = r"[0-9|\s]*/\s?[0-9|\s]*-?\s?[0-9|\s]*)[.|,]"
        orgao = r"Lotacao:|Quadro\sde\sPessoal\sd[a|e|o]([\s\S]*?)[.|,]"
        rules = {"processo_SEI": sei+sei2,
                 "nome": r"\s([^,]*?),\smatricula",
                 "matricula": r"matricula\s?n?o?\s([\s\S]*?)[,|\s]",
                 "tipo_ret": r"",
                 "cargo_efetivo": r"Cargo de([\s\S]*?)\,",
                 "classe": r"[C|c]lasse\s([\s\S]*?)\,",
                 "padrao": r"[p|P]adr[a|ã]o\s([\s\S]*?),",
                 "quadro": r"d?[e|a|o]?(Quadro[\s\S]*?)[,|;|.]",
                 "fundamento_legal": r"nos\stermos\sdo\s[a|A]rtigo([\s\S]*?),\sa?\s",
                 "orgao": orgao,
                 "vigencia": r"",
                 "matricula_SIAPE": siape}
        return rules

class RetAposentadoria(Atos):

    def __init__(self, file, backend):
        super().__init__(file, backend)

    def _regex_flags(self):
        return re.IGNORECASE

    def _act_name(self):
        return "Retificações de Aposentadoria"

    def _props_names(self):
        return ["Tipo do Ato", "Tipo de Documento", "Número do documento",
                "Data do documento ", "Número do DODF", "Data do DODF",
                "Página do DODF", "Nome do Servidor", "Matrícula", "Cargo",
                "Classe", "Padrao", "Matricula SIAPE",
                "Informação Errada", "Informação Corrigida"]

    def _rule_for_inst(self):
        start = r"(RETIFICAR,\s)"
        body = r"(.*?ato\sque\sconcedeu\saposentadoria[\s\S]*?"
        end = r"\.\n)"
        return start + body + end

    def _prop_rules(self):
        tipo = r"^n[a|o]\s([\s\S]*?),?\s?(?:[0-9]*?),"
        tipo2 = r"?\sde\s(?:[0-9]*?[/|.][0-9]*?[/|.][0-9]*?|,)"

        num_doc = r"n[a|o]\s(?:[\s\S]*?),?\s?([0-9]*?),"
        num_doc2 = r"?\sde\s(?:[0-9]*?[/|.][0-9]*?[/|.][0-9]*?|,)"

        data_doc = r"n[a|o]\s(?:[\s\S]*?),?\s?(?:[0-9]*?),?\sde\s"
        data_doc2 = r"([0-9]*?[/|.][0-9]*?[/|.][0-9]*?),\s"

        data_dodf = r"dodf[\s\S]*?(?:[0-9]*?)"
        data_dodf2 = r"([0-9]*?[/|.][0-9]*?[/|.][0-9]*?)[,|\s]"

        le = r"\sle[,|:|;]\s?([\s\S]*?),?\sleia[\s\S]*?"
        le2 = r"[,|:|;]\s(?:[\s\S]*?)[.]\s"

        leia = r"\sle[,|:|;]\s?(?:[\s\S]*?),"
        leia2 = r"?\sleia[\s\S]*?[,|:|;]\s([\s\S]*?)[.]\s"

        rules = {"tipo_documento": tipo + tipo2,
                 "numero_documento": num_doc + num_doc2,
                 "data_documento": data_doc + data_doc2,
                 "numero_dodf": r"dodf[\s\S]*?([0-9]*?),",
                 "data_dodf": data_dodf + data_dodf2,
                 "pagina_dodf": r"",
                 "nome": r"\sa\s([^,]*?),\smatricula",
                 "matricula": r"matricula\s?n?o?\s([\s\S]*?-[\s\S]*?)[,]",
                 "cargo_efetivo": r"(?:Cargo|Carreira)\sde([\s\S]*?)\,",
                 "classe": r"(?:([^,]*?)\sclasse,)?(?(1)|classe\s([\s\S]*?),)",
                 "padrao": r"[p|P]adr[a|ã]o\s([\s\S]*?),",
                 "matricula_SIAPE": r"siape\sn?o?\s([\s\S]*?)[,| | .]",
                 "informacao_errada": le+le2,
                 "informacao_corrigida": leia+leia2}
        return rules



"""Find titles using a Filter."""

import functools


class BoldUpperCase:
    """Filter functions useful for bold and upper case text.

    Note:
        This class is static and should not be instanciated.

    """
    TEXT_MIN = 4
    TRASH_WORDS = [
        "SUMÁRIO",
        "DIÁRIO OFICIAL",
        "SEÇÃO (I|II|III)",
    ]

    BOLD_FLAGS = [16, 20]

    @classmethod
    def dict_text(cls, d):
        """Check if text is title.

        Evaluates to true if d['text'] matches the following conditions:

            - all letters are uppercase;
            - does not contain 4 or more consecutive spaces;
            - has a len greater than BoldUpperCase.TEXT_MIN/

        Returns:
            Boolean indicating if text is title.
        """
        t = d['text'].strip().strip('.')
        cond1 = 4 * " " not in t
        cond2 = len(t) > cls.TEXT_MIN
        cond3 = t == t.upper()
        return cond1 and cond2 and cond3

    @classmethod
    def dict_bold(cls, d):
        """Hmm.

        Evaluates do True if d['flags'] matches the following conditions:

            - is one of the values in BoldUpperCase.BOLD_FLAGS

        """
        flags = d['flags']
        return flags in cls.BOLD_FLAGS



"""NER backend for act and propriety extraction.

This module contains the ActNER class, which have all that is necessary to
extract an act and, its proprieties, using a trained ner model.

"""

import numpy as np

class ActNER:
    """Act NER Class.

    This class encapsulate all functions, and attributes related
    to the process of NER extraction.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _model: The trained NER model for the act

    """

    def __init__(self):
        # self._backend = 'regex'
        super(ActNER, self).__init__()
        self._model = self._load_model()

    def _load_model(self):
        """Load Model from models/folder.

        Note:
            This function needs to be overwriten in
            the child class. If this function is not
            overwrite the backend will fall back to regex.

        """
        if self._backend == 'ner':
            print(f"Act {self._name} does not have an model: FALLING BACK TO REGEX")
            self._backend = 'regex'
        else:
            self._backend = 'regex'

    def _get_features(self, act):
        """Create features for each word in act.

        Create a list of dict of words features to be used in the predictor module.

        Args:
            act (list): List of words in an act.

        Returns:
            A list with a dictionary of features for each of the words.

        """
        sent_features = []
        for i in range(len(act)):
            word_feat = {
                'word': act[i].lower(),
                'capital_letter': act[i][0].isupper(),
                'all_capital': act[i].isupper(),
                'isdigit': act[i].isdigit(),
                'word_before': act[i].lower() if i == 0 else act[i-1].lower(),
                'word_after:': act[i].lower() if i+1 >= len(act) else act[i+1].lower(),
                'BOS': i == 0,
                'EOS': i == len(act)-1
            }
            sent_features.append(word_feat)
        return sent_features

    def _prediction(self, act):
        """Predict classes for a single act.

        Args:
            act (string): Full act

        Returns:
            A dictionary with the proprieties and its
            predicted value.
        """
        print("Predicting")
        act = self._preprocess(act)
        feats = self._get_features(act)
        predictions = self._model.predict_single(feats)
        return self._predictions_dict(act, predictions)

    def _preprocess(self, sentence):
        """Transform a raw string to a list of words.

        Args:
            sentence (string): Raw act as string.

        Returns:
            The list of words in the act.
        """
        sentence = sentence.replace(',', ' , ').replace(';', ' ; '). \
                   replace(':', ' : ').replace('. ', ' . ').replace('\n', ' ')
        if sentence[len(sentence)-2:] == '. ':
            sentence = sentence[:len(sentence)-2] + " ."
        return sentence.split()

    def _predictions_dict(self, sentence, prediction):
        """Create dictionary of proprieties.

        Create dictionary of tags to save predicted entities.

        Args:
            sentence (list): List of words and tokens in the act.
            prediction ([type]): The correspondent predicitons for each
                                 word in the sentence.

        Returns:
            A dictionary of the proprieties found.

        """
        tags = self._model.classes_
        tags.remove('O')
        tags.sort(reverse=True)
        while(tags[0][0] == 'I'):
            del tags[0]
        for i in range(len(tags)):
            tags[i] = tags[i][2:]
        dict = {}
        dict["Tipo do Ato"] = ""
        for i in tags:
            dict[i] = []

        last_tag = 'O'
        temp_entity = []
        for i in range(len(prediction)):
            if prediction[i] != last_tag and last_tag != 'O':
                dict[last_tag[2:]].append(temp_entity)
                temp_entity = []
            if prediction[i] == 'O':
                last_tag = 'O'
                continue
            else:
                temp_entity.append(sentence[i])
                last_tag = "I" + prediction[i][1:]

        if temp_entity:
            dict[last_tag[2:]].append(temp_entity)

        for key in dict.keys():
            values = []
            for value in dict[key]:
                value = ' '.join(value)
                values.append(value)
            if len(values) == 1:
                dict[key] = values[0]
            else:
                dict[key] = values


            if dict[key] == []:
                dict[key] = np.nan

        dict["Tipo do Ato"] = self._name


        return dict
